@@ -12,6 +12,13 @@
 #include <wchar.h>
 #include <ctype.h>
 #include <unistd.h>
+#include <errno.h>
+
+#define COLOR_WOOD_BRIGHT COLOR_PAIR(4)
+#define COLOR_WOOD_DARK COLOR_PAIR(2)
+#define COLOR_LEAF_BRIGHT COLOR_PAIR(3)
+#define COLOR_LEAF_DARK COLOR_PAIR(1)
+#define COLOR_TEXT COLOR_PAIR(5)
 
 enum branchType {trunk, shootLeft, shootRight, dying, dead};
 
@@ -31,10 +38,11 @@ struct config {
 	int targetBranchCount;
 
 	double timeWait;
-	double timeStep;
+	float timeStep;
 
 	char* message;
 	char* leaves[64];
+	int colors[4];
 	char* saveFile;
 	char* loadFile;
 };
@@ -57,7 +65,7 @@ struct counters {
 	int shootCounter;
 };
 
-void quit(struct config *conf, struct ncursesObjects *objects, int returnCode) {
+void delObjects(struct ncursesObjects *objects) {
 	// delete panels
 	del_panel(objects->basePanel);
 	del_panel(objects->treePanel);
@@ -69,7 +77,10 @@ void quit(struct config *conf, struct ncursesObjects *objects, int returnCode) {
 	delwin(objects->treeWin);
 	delwin(objects->messageBorderWin);
 	delwin(objects->messageWin);
+}
 
+void quit(struct config *conf, struct ncursesObjects *objects, int returnCode) {
+	delObjects(objects);
 	free(conf->saveFile);
 	free(conf->loadFile);
 	exit(returnCode);
@@ -121,47 +132,52 @@ void finish(const struct config *conf, struct counters *myCounters) {
 }
 
 void printHelp(void) {
-	printf("Usage: cbonsai [OPTION]...\n");
-	printf("\n");
-	printf("cbonsai is a beautifully random bonsai tree generator.\n");
-	printf("\n");
-	printf("Options:\n");
-	printf("  -l, --live             live mode: show each step of growth\n");
-	printf("  -t, --time=TIME        in live mode, wait TIME secs between\n");
-	printf("                           steps of growth (must be larger than 0) [default: 0.03]\n");
-	printf("  -i, --infinite         infinite mode: keep growing trees\n");
-	printf("  -w, --wait=TIME        in infinite mode, wait TIME between each tree\n");
-	printf("                           generation [default: 4.00]\n");
-	printf("  -S, --screensaver      screensaver mode; equivalent to -li and\n");
-	printf("                           quit on any keypress\n");
-	printf("  -m, --message=STR      attach message next to the tree\n");
-	printf("  -b, --base=INT         ascii-art plant base to use, 0 is none\n");
-	printf("  -c, --leaf=LIST        list of comma-delimited strings randomly chosen\n");
-	printf("                           for leaves\n");
-	printf("  -M, --multiplier=INT   branch multiplier; higher -> more\n");
-	printf("                           branching (0-20) [default: 5]\n");
-	printf("  -L, --life=INT         life; higher -> more growth (0-200) [default: 32]\n");
-	printf("  -p, --print            print tree to terminal when finished\n");
-	printf("  -s, --seed=INT         seed random number generator\n");
-	printf("  -W, --save=FILE        save progress to file [default: $XDG_CACHE_HOME/cbonsai or $HOME/.cache/cbonsai]\n");
-	printf("  -C, --load=FILE        load progress from file [default: $XDG_CACHE_HOME/cbonsai]\n");
-	printf("  -v, --verbose          increase output verbosity\n");
-	printf("  -h, --help             show help	\n");
+	printf("%s",
+		"Usage: cbonsai [OPTION]...\n"
+	        "\n"
+	        "cbonsai is a beautifully random bonsai tree generator.\n"
+	        "\n"
+	        "Options:\n"
+	        "  -l, --live             live mode: show each step of growth\n"
+	        "  -t, --time=TIME        in live mode, wait TIME secs between\n"
+	        "                           steps of growth (must be larger than 0) [default: 0.03]\n"
+	        "  -i, --infinite         infinite mode: keep growing trees\n"
+	        "  -w, --wait=TIME        in infinite mode, wait TIME between each tree\n"
+	        "                           generation [default: 4.00]\n"
+	        "  -S, --screensaver      screensaver mode; equivalent to -li and\n"
+	        "                           quit on any keypress\n"
+	        "  -m, --message=STR      attach message next to the tree\n"
+	        "  -b, --base=INT         ascii-art plant base to use, 0 is none\n"
+	        "  -c, --leaf=LIST        list of comma-delimited strings randomly chosen\n"
+	        "                           for leaves [default: &]\n"
+	        "  -k, --color=LIST       list of 4 comma-delimited color indices (0-255) for\n"
+	        "                           each of dark leaves, dark wood, light leaves, and\n"
+	        "                           light wood, in that order [default: 2,3,10,11]\n"
+	        "  -M, --multiplier=INT   branch multiplier; higher -> more\n"
+	        "                           branching (0-20) [default: 5]\n"
+	        "  -L, --life=INT         life; higher -> more growth (0-200) [default: 32]\n"
+	        "  -p, --print            print tree to terminal when finished\n"
+	        "  -s, --seed=INT         seed random number generator\n"
+	        "  -W, --save=FILE        save progress to file [default: $XDG_CACHE_HOME/cbonsai or $HOME/.cache/cbonsai]\n"
+	        "  -C, --load=FILE        load progress from file [default: $XDG_CACHE_HOME/cbonsai]\n"
+	        "  -v, --verbose          increase output verbosity\n"
+	        "  -h, --help             show help\n"
+    );
 }
 
 void drawBase(WINDOW* baseWin, int baseType) {
 	// draw base art
 	switch(baseType) {
 	case 1:
-		wattron(baseWin, A_BOLD | COLOR_PAIR(8));
+		wattron(baseWin, A_BOLD | COLOR_TEXT);
 		wprintw(baseWin, "%s", ":");
-		wattron(baseWin, COLOR_PAIR(2));
+		wattron(baseWin, COLOR_LEAF_BRIGHT);
 		wprintw(baseWin, "%s", "___________");
-		wattron(baseWin, COLOR_PAIR(11));
+		wattron(baseWin, COLOR_WOOD_BRIGHT);
 		wprintw(baseWin, "%s", "./~~~\\.");
-		wattron(baseWin, COLOR_PAIR(2));
+		wattron(baseWin, COLOR_LEAF_BRIGHT);
 		wprintw(baseWin, "%s", "___________");
-		wattron(baseWin, COLOR_PAIR(8));
+		wattron(baseWin, COLOR_TEXT);
 		wprintw(baseWin, "%s", ":");
 
 		mvwprintw(baseWin, 1, 0, "%s", " \\                           / ");
@@ -171,15 +187,15 @@ void drawBase(WINDOW* baseWin, int baseType) {
 		wattroff(baseWin, A_BOLD);
 		break;
 	case 2:
-		wattron(baseWin, COLOR_PAIR(8));
+		wattron(baseWin, COLOR_TEXT);
 		wprintw(baseWin, "%s", "(");
-		wattron(baseWin, COLOR_PAIR(2));
+		wattron(baseWin, COLOR_LEAF_BRIGHT);
 		wprintw(baseWin, "%s", "---");
-		wattron(baseWin, COLOR_PAIR(11));
+		wattron(baseWin, COLOR_WOOD_BRIGHT);
 		wprintw(baseWin, "%s", "./~~~\\.");
-		wattron(baseWin, COLOR_PAIR(2));
+		wattron(baseWin, COLOR_LEAF_BRIGHT);
 		wprintw(baseWin, "%s", "---");
-		wattron(baseWin, COLOR_PAIR(8));
+		wattron(baseWin, COLOR_TEXT);
 		wprintw(baseWin, "%s", ")");
 
 		mvwprintw(baseWin, 1, 0, "%s", " (           ) ");
@@ -209,20 +225,16 @@ void drawWins(int baseType, struct ncursesObjects *objects) {
 	int baseOriginY = (rows - baseHeight);
 	int baseOriginX = (cols / 2) - (baseWidth / 2);
 
+	// clean up old objects
+	delObjects(objects);
+
 	// create windows
 	objects->baseWin = newwin(baseHeight, baseWidth, baseOriginY, baseOriginX);
 	objects->treeWin = newwin(rows - baseHeight, cols, 0, 0);
 
-	// create/replace tree and base panels
-	if (objects->basePanel)
-		replace_panel(objects->basePanel, objects->baseWin);
-	else
-		objects->basePanel = new_panel(objects->baseWin);
-
-	if (objects->treePanel)
-		replace_panel(objects->treePanel, objects->treeWin);
-	else
-		objects->treePanel = new_panel(objects->treeWin);
+	// create tree and base panels
+	objects->basePanel = new_panel(objects->baseWin);
+	objects->treePanel = new_panel(objects->treeWin);
 
 	drawBase(objects->baseWin, baseType);
 }
@@ -257,18 +269,18 @@ void chooseColor(enum branchType type, WINDOW* treeWin) {
 	case trunk:
 	case shootLeft:
 	case shootRight:
-		if (rand() % 2 == 0) wattron(treeWin, A_BOLD | COLOR_PAIR(11));
-		else wattron(treeWin, COLOR_PAIR(3));
+		if (rand() % 2 == 0) wattron(treeWin, A_BOLD | COLOR_WOOD_BRIGHT);
+		else wattron(treeWin, COLOR_WOOD_DARK);
 		break;
 
 	case dying:
-		if (rand() % 10 == 0) wattron(treeWin, A_BOLD | COLOR_PAIR(2));
-		else wattron(treeWin, COLOR_PAIR(2));
+		if (rand() % 10 == 0) wattron(treeWin, A_BOLD | COLOR_LEAF_BRIGHT);
+		else wattron(treeWin, COLOR_LEAF_BRIGHT);
 		break;
 
 	case dead:
-		if (rand() % 3 == 0) wattron(treeWin, A_BOLD | COLOR_PAIR(10));
-		else wattron(treeWin, COLOR_PAIR(10));
+		if (rand() % 3 == 0) wattron(treeWin, A_BOLD | COLOR_LEAF_DARK);
+		else wattron(treeWin, COLOR_LEAF_DARK);
 		break;
 	}
 }
@@ -534,19 +546,12 @@ void createMessageWindows(struct ncursesObjects *objects, char* message) {
 	objects->messageWin = newwin(boxHeight, boxWidth + 1, maxY * 0.7, maxX * 0.7);
 
 	// draw box
-	wattron(objects->messageBorderWin, COLOR_PAIR(8) | A_BOLD);
+	wattron(objects->messageBorderWin, COLOR_TEXT | A_BOLD);
 	wborder(objects->messageBorderWin, '|', '|', '-', '-', '+', '+', '+', '+');
 
-	// create/replace message panels
-	if (objects->messageBorderPanel)
-		replace_panel(objects->messageBorderPanel, objects->messageBorderWin);
-	else
-		objects->messageBorderPanel = new_panel(objects->messageBorderWin);
-
-	if (objects->messagePanel)
-		replace_panel(objects->messagePanel, objects->messageWin);
-	else
-		objects->messagePanel = new_panel(objects->messageWin);
+	// create message panels
+	objects->messageBorderPanel = new_panel(objects->messageBorderWin);
+	objects->messagePanel = new_panel(objects->messageWin);
 }
 
 int drawMessage(const struct config *conf, struct ncursesObjects *objects, char* message) {
@@ -629,7 +634,7 @@ int drawMessage(const struct config *conf, struct ncursesObjects *objects, char*
 
 		if (conf->verbosity >= 2) {
 			updateScreen(1);
-			mvwprintw(objects->treeWin, 11, 5, "word buffer: |% 15s|", wordBuffer);
+			mvwprintw(objects->treeWin, 11, 5, "word buffer: |%-15s|", wordBuffer);
 		}
 		if (thisChar == '\0') break;	// quit when we reach the end of the message
 		i++;
@@ -653,21 +658,32 @@ void init(const struct config *conf, struct ncursesObjects *objects) {
 		int bg = COLOR_BLACK;
 		if (use_default_colors() != ERR) bg = -1;
 
-		// define color pairs
-		for(int i=0; i<16; i++){
-			init_pair(i, i, bg);
+		// initialize color pairs
+		int warnedAlready = 0;
+		for(int id = 1; id <= 4; id++) {
+			int color_id = conf->colors[id-1];
+
+			// restrict color pallete in non-256color terminals
+			if (COLORS < 256) {
+				init_pair(id, color_id % 8, bg);
+			} else {
+				init_pair(id, color_id, bg);
+			}
+
+			if (!warnedAlready && COLORS < 256 && color_id >= 8) {
+				endwin();
+				printf("Warning: defaulting to 8-color support.\n");
+				warnedAlready = 1;
+				doupdate();
+			}
 		}
 
-		// restrict color pallete in non-256color terminals (e.g. screen or linux)
+		// initialize COLOR_TEXT pair
 		if (COLORS < 256) {
-			init_pair(8, 7, bg);	// gray will look white
-			init_pair(9, 1, bg);
-			init_pair(10, 2, bg);
-			init_pair(11, 3, bg);
-			init_pair(12, 4, bg);
-			init_pair(13, 5, bg);
-			init_pair(14, 6, bg);
-			init_pair(15, 7, bg);
+			init_pair(5, 7, bg);
+		}
+		else {
+			init_pair(5, 8, bg);
 		}
 	} else {
 		printf("%s", "Warning: terminal does not have color support.\n");
@@ -727,11 +743,17 @@ void printstdscr(void) {
 
 			// enable correct color
 			if (fg == 0) printf("\033[0m");
+			else if (fg >= 16) printf("\033[38;5;%him", fg);
 			else if (fg <= 7) printf("\033[3%him", fg);
 			else if (fg >= 8) printf("\033[9%him", fg - 8);
 
+			// print wchar
 			printf("%ls", wch);
 
+			// if we're on the last char in the row, print a newline
+			if (x == maxX - 1)
+				printf("\n");
+
 			short clen = wcslen(wch);
 			short cwidth = 0;
 			for (int i = 0; i < clen; ++i)
@@ -804,6 +826,7 @@ int main(int argc, char* argv[]) {
 
 		.message = NULL,
 		.leaves = {0},
+		.colors = {0, 0, 0, 0},
 		.saveFile = createDefaultCachePath(),
 		.loadFile = createDefaultCachePath(),
 	};
@@ -817,6 +840,7 @@ int main(int argc, char* argv[]) {
 		{"message", required_argument, NULL, 'm'},
 		{"base", required_argument, NULL, 'b'},
 		{"leaf", required_argument, NULL, 'c'},
+		{"colors", required_argument, NULL, 'k'},
 		{"multiplier", required_argument, NULL, 'M'},
 		{"life", required_argument, NULL, 'L'},
 		{"print", required_argument, NULL, 'p'},
@@ -831,17 +855,18 @@ int main(int argc, char* argv[]) {
 	struct ncursesObjects objects = { NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL };
 
 	char leavesInput[128] = "&";
+	char colorsInput[128] = "2,3,10,11";
 
 	// parse arguments
 	int option_index = 0;
 	int c;
-	while ((c = getopt_long(argc, argv, ":lt:iw:Sm:b:c:M:L:ps:C:W:vh", long_options, &option_index)) != -1) {
+	while ((c = getopt_long(argc, argv, ":lt:iw:Sm:b:c:k:M:L:ps:C:W:vh", long_options, &option_index)) != -1) {
 		switch (c) {
 		case 'l':
 			conf.live = 1;
 			break;
 		case 't':
-			if (strtold(optarg, NULL) != 0) conf.timeStep = strtod(optarg, NULL);
+			if (strtof(optarg, NULL) != 0) conf.timeStep = strtof(optarg, NULL);
 			else {
 				printf("error: invalid step time: '%s'\n", optarg);
 				quit(&conf, &objects, 1);
@@ -878,7 +903,13 @@ int main(int argc, char* argv[]) {
 			conf.message = optarg;
 			break;
 		case 'b':
-			if (strtold(optarg, NULL) != 0) conf.baseType = strtod(optarg, NULL);
+                        /* 0 can legitimately be returned, so we cannot check wether
+                           strtold(optarg, NULL) != 0.  We need to set errno to zero
+                           before the conversion attempt, and check it it changed
+                           afterwards. */
+                        errno = 0;
+                        strtold(optarg, NULL);
+                        if (!errno) conf.baseType = strtod(optarg, NULL);
 			else {
 				printf("error: invalid base index: '%s'\n", optarg);
 				quit(&conf, &objects, 1);
@@ -888,6 +919,10 @@ int main(int argc, char* argv[]) {
 			strncpy(leavesInput, optarg, sizeof(leavesInput) - 1);
 			leavesInput[sizeof(leavesInput) - 1] = '\0';
 			break;
+		case 'k':
+			strncpy(colorsInput, optarg, sizeof(colorsInput) - 1);
+			colorsInput[sizeof(colorsInput) - 1] = '\0';
+			break;
 		case 'M':
 			if (strtold(optarg, NULL) != 0) conf.multiplier = strtod(optarg, NULL);
 			else {
@@ -993,6 +1028,32 @@ int main(int argc, char* argv[]) {
 		conf.leavesSize++;
 	}
 
+	// delimit colors on "," and add each color to the colors[] list
+	token = strtok(colorsInput, ",");
+	for(int i = 0; i<4; i++) {
+		if (token == NULL) {
+			printf("error: too few color indices provided\n");
+			quit(&conf, &objects, 1);
+		}
+		char* temp;
+		errno = 0;
+
+		// parse each number
+		long parsed = strtol(token, &temp, 0);
+		if (temp == token || *temp != '\0' || errno == ERANGE || parsed < 0 || parsed >= 256) {
+			printf("error: invalid color index: '%s'\n", token);
+			quit(&conf, &objects, 1);
+		} else {
+			// add to list and continue
+			conf.colors[i] = parsed;
+			token = strtok(NULL, ",");
+		}
+	}
+	if (token != NULL) {
+		printf("error: too many color indices provided\n");
+		quit(&conf, &objects, 1);
+	}
+
 	if (conf.load)
 		loadFromFile(&conf);
 
